﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

[RequireComponent(typeof(CharacterController))]
public class Movement : MonoBehaviour
{
    // The Speed at which the Player can move in units per second
    [Tooltip("How fast the character moves when walking (units per second).")]
    [SerializeField] float moveSpeed = 4f;

    // Allow Jumping
    [Tooltip("Should Jumping be allowed?")]
    [SerializeField] bool allowJump = true;

    // The height of a jump, in units
    [Tooltip("How high should the character jump (in units).")]
    [SerializeField] float jumpHeight = 1.5f;

    // The rate at which our vertical speed will be reduced in units per second
    [Tooltip("Gravity in units per second.")]
    [SerializeField] float gravity = 30f;

    // The degree to which movement can be controlled midair
    [Tooltip("How much maneuverability should remain in the air.")]
    [Range(0, 10), SerializeField] float airControl = 5;

    // Movement Direction, can be only controlled on the ground
    Vector3 moveDirection = Vector3.zero;

    // Reference to the Character Controller (cached)
    CharacterController controller;

    // Start is called before the first frame update
    void Start()
    {
        controller = GetComponent<CharacterController>();
    }

    // FixedUpdate must be used if the Player should interact with Physics Objects
    void FixedUpdate()
    {
        // This vector describes the desired plane of movement. If we are in the
        // air we will interpolate the current movement with this vector to
        // simulate a form of momentum
        var input = new Vector3(
            Input.GetAxis("Horizontal"),
            0,
            Input.GetAxis("Vertical")
        );

        // Multiply the movement with the movement speed
        input *= moveSpeed;

        // Convert the rotation to world-space coordinates (as needed by 
        // the CharacterController Type)
        input = transform.TransformDirection(input);

        // Is the Controllers bottom-most point touching the Ground?
        if (controller.isGrounded)
        {
            moveDirection = input;
            // If jumping is allowed and the Jump Button is pressed
            if (allowJump && Input.GetButton("Jump"))
            {
                moveDirection.y = Mathf.Sqrt(2 * gravity * jumpHeight);
            }
            else
            {
                // Player is on the Ground, cancel Y-Axis Movement
                moveDirection.y = 0;
            }
        }
        else
        {
            // We are falling, so Ignore Y-Axis-Input, but allow user to steer
            // the fall on the X/Z-Axis
            input.y = moveDirection.y;
            moveDirection = Vector3.Lerp(moveDirection, input, airControl * Time.deltaTime);
        }
        // Subtract the gravity from any Y-Axis movement
        moveDirection.y -= gravity * Time.deltaTime;

        // Finally: Move the controller into the desired direction.
        // Note: if this would hit a collider or the controller would refuse
        //       to move further – however – if the other collider would move
        //       into this object it won't budge..
        controller.Move(moveDirection * Time.deltaTime);
    }
}
